#!/usr/bin/python
import nrrd
import json
import os
import numpy as np
from collections import defaultdict
output_img_size = [528, 456, 320]

use_aibs_ref = True
env_path = "/home/ubuntu/data/"
# change these paths as need be for environemnt
base_path = os.path.join(env_path,"cellpos") # base path for cell positions
aibs_ref = os.path.join(env_path,"aibsref/aibsAnnot25-transformed.nrrd") # path to nrrd of annotated 
man_ref = os.path.join(env_path,"manref/manref.nrrd") # optional path of manual reference annotation
cell_pos = os.path.join(env_path,"cellpos/warped_csv_img") # path to directory of nrrd files with cell count volumes
cell_count_out = os.path.join(env_path,"cellpos/cell_count") # path to 
annot_path = os.path.join(env_path,"aibsref/atlas_annotations_trimmed.json") # annotations for ABA
man_annot_path = os.path.join(env_path,"manref/annot.txt") # annotations for optional manual reference
expt_path = os.path.join(env_path,"cellpos/types.txt") # types of each experiment (speicifc to Ye et al 2016)

from joblib import Parallel,delayed


def load_expt_types(fname):
    expts = defaultdict(list)
    with open(fname,'r') as f:
        for l in f:
            print l
            fields = l.rstrip().split(" ")
            if fields[1] == "h":
                expts['h'].append(int(fields[0]))
            elif fields[1] == "c":
                expts['c'].append(int(fields[0]))
            elif fields[1] == "s":
                expts['s'].append(int(fields[0]))
    return expts

def load_annot(fname):
    f = open(fname,'r').read()
    elems = f.split("\n")

    regions = {}
    parents = defaultdict(list)
    for k in range(0, len(elems), 4):
        if len(elems[k]) > 0:
            id = int(elems[k].split(":")[1].rstrip())
            acronym = elems[k+1].split(":")[1].rstrip()
            name = elems[k+1].split(":")
            parent = int(elems[k+3].split(":")[1].rstrip())
            regions[id] = (acronym, name, parent)
            parents[parent].append((id, acronym, name))
    return (regions, parents)

def count_vol(fname, annot_vol, annot):
    vol = nrrd.read(fname)[0]
    counts = {}
    for k in annot.keys():
        curridx = np.where(annot_vol == int(k))
        counts[k] = sum(vol[curridx])
#        print k, annot[k][0], counts[k]
    return counts

def count_cells_to_file(f, annot_vol, annot):
    mouse_name = f.split(".")[0]
    path = os.path.join(cell_pos, f)
    print "Counting cells", f
    counts = count_vol(path, annot_vol, annot)
    fout = open(os.path.join(cell_count_out, mouse_name + "_count.txt"),'w')
    for k,v in counts.iteritems():
        fout.write("%d %s %d\n" % (k, annot[k][0], v))
    fout.close()

def make_avg_count_vols(expts):
    h = np.zeros(output_img_size); nh = 0
    s = np.zeros(output_img_size); ns = 0
    c = np.zeros(output_img_size); nc = 0
    for k,v in expts.iteritems():
        for i in v:
            print k,i
            fname = os.path.join(cell_pos, str(i) + ".nrrd")
            if os.path.exists(fname):
                vol = nrrd.read(fname)
                if k == "h":
                    h += vol[0]; nh += 1
                elif k == "s":
                    s += vol[0]; ns += 1
                elif k == "c":
                    c += vol[0]; nc += 1
    nrrd.write(os.path.join(base_path, "merged_vols", "h.nrrd"), h/float(nh))
    nrrd.write(os.path.join(base_path, "merged_vols", "s.nrrd"), s/float(ns))
    nrrd.write(os.path.join(base_path, "merged_vols", "c.nrrd"), c/float(nc))

def load_man_annot(path):
    regions = {}
    with open(path,'r') as f:
        for l in f:
            fields = l.rstrip().split(" ")
            regions[int(fields[1])] = (fields[2],)
    return regions

def aggregate_counts(dirname, expts, outfile):
    
    fout = open(outfile,'w')
    fout.write("regionID regionName count exptType exptID\n")
    for key, vals in expts.iteritems():
        for v in vals:
            curr_fname = os.path.join(dirname, str(v) + "_count.txt")
            if os.path.exists(curr_fname):
                with open(curr_fname,'r') as f:
                    for l in f:
                        curr = l.rstrip() + " " + key + " " + str(v)
                        fout.write(curr + "\n")
    fout.close()


                
def main():
    print "Loading annot"
    if use_aibs_ref:
        annot_vol = nrrd.read(aibs_ref)[0]
    else:
        annot_vol = nrrd.read(man_ref)[0]
    print annot_vol.shape
    if use_aibs_ref:
        (annot, parents) = load_annot(annot_path)
    else:
        annot = load_man_annot(man_annot_path)
    expts = load_expt_types(expt_path)
    if not os.path.exists(cell_count_out):
        os.mkdir(cell_count_out)
    #        gl_volumes = Parallel(n_jobs=10, verbose=50)(delayed(gaussian_laplace)(volume, s) for s in sigma_list)
    Parallel(n_jobs=16, verbose=50)(delayed(count_cells_to_file)(f, annot_vol, annot) for f in os.listdir(cell_pos))
    
    make_avg_count_vols(expts)
    aggregate_counts(cell_count_out, expts, os.path.join(base_path, "allcounts.txt"))

if __name__ == "__main__":
    main()
