#/usr/bin/python
import numpy as np
from PIL import Image
import os
import re
import sys
from scipy import misc
import scipy.ndimage
from skimage.transform import downscale_local_mean
import multiprocessing
from joblib import Parallel, delayed
import matplotlib.pyplot as plt
out_dir = "/data/processed_clarity/trap_downsamp_tiff/" # output location of downsampled TIFFs
base_dir = "/data/processed_clarity/raw_trap_data/TRAPs" # location of raw TIFF files

def get_fnames_refonly(dirname):
    fnames = sorted(os.listdir(dirname))
    names = []
    for f in fnames:
        if not "xml" in f:
            names.append(f)
    return names
 
def get_fnames(dirname):
    """ Get filenames for each channel """
    fnames = sorted(os.listdir(dirname))
    ch1_names = []
    ch2_names = []
    for f in fnames:
        if "Filter0001" in f and not "xml" in f:
            ch1_names.append(f)
        elif "Filter0000" in f and not "xml" in f:
            ch2_names.append(f) 
    return (ch1_names, ch2_names)

def load_vol(dirname, fnames, scale):
    """ Load volume from list of fnames in dirname, and scales each image by scale in X and Y (not Z) """
    if len(fnames) > 0:
        imgs = []
        im = misc.imread(os.path.join(dirname, fnames[0]))
        n = len(fnames)
        print "Allocating memory..."
        imgs = np.zeros((int(im.shape[0]/float(1)),
                         int(im.shape[1]/float(1)),
                         n), dtype=np.uint16)
        print "Loading data..."
        num_cores = multiprocessing.cpu_count()
        imlist = Parallel(n_jobs=num_cores)(delayed(_load_and_resize)(os.path.join(dirname, fnames[i]), scale) for i in xrange(n))
        print "Copying data..."
    #    imgs = np.array(imlist,dtype=np.uint16)
        for i in xrange(n):
            print "\r>> ",i/float(n),
            imgs[:,:,i] = imlist[i]
        print
        print imgs.shape
        return imgs
    else:
        return []


def _load_and_resize(fname, scale):
    img = plt.imread(fname).astype(np.uint16)
    return img
    #return downscale_local_mean(img, (scale, scale)) #misc.imresize(img, 1./scale, interp="bilinear")

def resize_z(img, scale): 
    downscaled = downscale_local_mean(img, (scale,scale,scale))
    downscaled /= downscaled.max()
    return (downscaled * 65535).astype(np.uint16) # convert to 16 bits

def save_volume_tiff_stack(fname, volume):
    from tifffile import imsave
    imsave(fname, volume) 
   
def downsamp_vol(dirname, fnames, out_dir, scale, out_ext):
    print "Loading data..."
    ch1 = load_vol(dirname, fnames, scale)
    if len(ch1) > 0:
        print ch1.shape
        print "Resizing..."
        ch1 = resize_z(ch1, scale)
        print ch1.shape
        print "Saving out..."
        save_volume_tiff_stack(out_dir + "_" + out_ext + ".tif", ch1)
        del ch1


def main():
    scale = 4 
    ref_only = False
    already_processed = os.listdir(out_dir)
    print already_processed
    for expt_name in os.listdir(base_dir):
    #expt_name = "150520_329_08iso_2sh144NA5sp200s90p_10-18-49"
        dirname = os.path.join(base_dir,expt_name)
        print "Processing:", dirname 
        if "352" in expt_name:
            if ref_only:
                fnames = get_fnames_refonly(dirname)
                downsamp_vol(dirname, fnames, out_dir+expt_name, scale, "ch2")
            else:
                (ch1_names, ch2_names) = get_fnames(dirname)
                downsamp_vol(dirname, ch1_names, out_dir+expt_name, scale, "ch1")
                downsamp_vol(dirname, ch2_names, out_dir+expt_name, scale, "ch2")

if __name__ == "__main__":
    main()
