#!/usr/bin/python
# Created by: William E. Allen (wallen@stanford.edu), July 2015
from register_elastix import *

from subprocess import PIPE, Popen
import os
import subprocess
import SimpleITK as sitk
import numpy as np
from util import *
import numpy as np
import nrrd

# input directories -- change these for particular environment
env_path = "/home/ubuntu/data"
base_dir = os.path.join("registered_elastix") # base directory for everything
data_dir = os.path.join("trap_downsamp_nii") # directory containing all nii files (ch1 and ch2) for ref building and registration
aibs_atlas = os.path.join("allenAtlas.nrrd") # ABA volume used to initialize process
cell_dir = os.path.join("cellpos") # locations of the cells to be warped

# output directories -- created by program -- don't edit
reg_dir = os.path.join(base_dir, "reg") # output dir -- 
final_out_dir = os.path.join(base_dir, "output")
out_dir = os.path.join(base_dir, "warped")
inverse_dir = os.path.join(base_dir, "inverse")

# constants of image size (based on LaVision LSM size and ABA size)
input_img_size = [640, 540, 309]
output_img_size = [528, 456, 320]

fnames = os.listdir(data_dir) # filenames to be included


def elastix_register(ref, floating, output_dir):
    cmd = "elastix -f %s -m %s -out %s -p params/rigid.txt -p params/affine.txt -p params/bspline.txt > /tmp/out.txt 2>&1" % (ref, floating, output_dir)
    print cmd
    os.system(cmd)   

def bootstrap_reg(fnames):
    ref = aibs_atlas
    k = 1
    for f in fnames:
        print "-->\t[",k,"/",len(fnames),"]:", f
        fname = os.path.join(data_dir, f)
        out_dir = os.path.join(reg_dir, "reg_" + f)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        print "-->\t Changing to", out_dir
        cmd = "elastix -f %s -m %s -out %s -p params/rigid.txt > /tmp/out.txt 2>&1" % (ref, fname, out_dir)
        os.system(cmd)
        k += 1
    # average together
    rigid_fnames = [os.path.join(reg_dir, "reg_"+f, "result.0.nrrd") for f in fnames]
    average_volumes_old(rigid_fnames, os.path.join(final_out_dir, "bootstrap_ref.nrrd"))

def iterative_affine_reg(fnames):
    niter = 1
    for i in xrange(1,niter):
        print "-->\tAFFINE ITERATION", i+1, "/", niter
        if i == 0:
            ref_name = "bootstrap_ref.nrrd"
        else:
            ref_name = "affine_ref.nrrd" 
        ref = os.path.join(final_out_dir, ref_name)
        k = 1
        for f in fnames:
            print "-->\t[",k,"/",len(fnames),"]:", f
            fname = os.path.join(data_dir, f)
            out_dir = os.path.join(reg_dir, "reg_" + f)
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)
            print "-->\t Changing to", out_dir
            cmd = "elastix -f %s -m %s -out %s -p params/rigid.txt -p params/affine.txt > /tmp/out.txt 2>&1" % (ref, fname, out_dir)
            #os.system(cmd)
            k += 1
        next_ref_name = "affine_ref.nrrd" 
        affine_fnames = [os.path.join(reg_dir, "reg_"+f, "result.1.nrrd") for f in fnames]
        average_volumes_old(affine_fnames, os.path.join(final_out_dir, next_ref_name))

def iterative_warp_reg(fnames, restart=True):
    niter = 5
    for i in xrange(niter):
        print "-->\tWARP ITERATION", i+1, "/", niter
        if i == 0 and restart:
            ref_name = "affine_ref.nrrd"
        else: 
            ref_name = "warp_ref.nrrd" 
        print "Using as ref:", ref_name
        ref = os.path.join(final_out_dir, ref_name)
        k = 1
        for f in fnames:
            print "-->\t[",k,"/",len(fnames),"]:", f
            fname = os.path.join(data_dir, f)
            out_dir = os.path.join(reg_dir, "reg_" + f)
            if not os.path.exists(out_dir):
                os.makedirs(out_dir)
            print "-->\t Changing to", out_dir
            cmd = "elastix -f %s -m %s -out %s -p params/rigid.txt -p params/affine.txt -p params/bspline.txt > /tmp/out.txt 2>&1" % (ref, fname, out_dir)
            os.system(cmd)
            k += 1
        next_ref_name = "warp_ref.nrrd" 
        warp_fnames = [os.path.join(reg_dir, "reg_"+f, "result.2.nrrd") for f in fnames]
        average_volumes_old(warp_fnames, os.path.join(final_out_dir, next_ref_name))

def apply_reg_all(ref_name, fnames):
    print fnames
    for i in xrange(len(fnames)):
        f = fnames[i]
        if not "flip" in f and "ch1" in f:  
            os.chdir(os.path.join(reg_dir, "reg_"+f.replace("ch1","ch2")))
            reg_name = "TransformParameters.2.txt"
            float_name = os.path.join(data_dir, f)
            out_name = os.path.join(out_dir, "reg_" + f)
            #os.mkdir(out_name)
            print "TRANFORMING", float_name, out_name
            apply_reg(ref_name, float_name, out_name, reg_name)

def apply_reg(ref_name, float_name, out_name, reg_name):
# TransformParameters.2.txt
    cmd = "transformix -in %s -out %s -tp %s" % (float_name, out_name, reg_name)
    print cmd
    os.system(cmd)

def downsamp_csv(csv_fname, out_fname):
    """ Downsample CSV file from imaris to image coords """
    f = open(csv_fname,'r')
    k = 0
    cellpos = []
    c = 1./(0.025*1000) # conversion factor
    #ref_pix_scale = (4*5.16)/25 # convert to fixed image domain
    #c = ref_pix_scale / 1000 # fixed image domain
    for l in f:
        if k > 4:
            fields = l.rstrip().split(",")
            cellpos.append((float(fields[1])*c, float(fields[0])*c, float(fields[2])*c)) # note reversal of x,y
        k += 1
    f.close()
    fout = open(out_fname, 'w') # write in format for elastix
    fout.write("index\n")
    fout.write("%d\n" % len(cellpos))
    for cell in cellpos:
        fout.write("%d %d %d\n" % (round(cell[0]),round(cell[1]),round(cell[2]))) 
    fout.close()
    return cellpos

def warp_points(pts_fname, out_dir, transform):
    cmd = "transformix -def %s -out %s -tp %s" % (pts_fname, out_dir, transform)
    print cmd
    os.system(cmd)

def warp_all_points(csv_dir, downsamp_dir, warped_dir):
    options = {}
    options['spacings'] = [pixel_size, pixel_size, pixel_size]
    for f in os.listdir(csv_dir):
        mouse_name = f.split(".")[0] 
        reg_name = os.path.join(inverse_dir, "reg_" + mouse_name + "_ch2.nii", "TransformParameters.0.txt") #
        downsamp_name = os.path.join(downsamp_dir, f)
        warped_out = os.path.join(warped_dir, "elastix_out", f)
        print "Downsampling ", f
        cellpos = downsamp_csv(os.path.join(csv_dir, f), downsamp_name)
        nrrd.write(os.path.join(cell_dir, "csv_img", mouse_name + ".nrrd"), make_count_image(cellpos, input_img_size), options)
        print "Warping", f
        if not os.path.exists(warped_out):
            os.mkdir(warped_out)
        if os.path.exists(reg_name):
            rewrite_inv_file(reg_name) # edit inverse file to just apply inverse
            warp_points(downsamp_name, warped_out, reg_name) # apply warp
        # convert to useful csv format from weird elastix format
            elastix_pts = os.path.join(warped_out, "outputpoints.txt")

            (inputidx, idx, pts) = parse_warped_points(elastix_pts)

            nrrd.write(os.path.join(cell_dir, "warped_csv_img", mouse_name + ".nrrd"), make_count_image(idx, output_img_size),  options)
            f = open(os.path.join(warped_dir, mouse_name + "_warped_idx.csv"),'w')
            for pos in idx:
                f.write("%d %d %d\n" % (pos[0], pos[1], pos[2]))
            f.close()
            f = open(os.path.join(warped_dir, mouse_name + "_warped_pt.csv"),'w')
            for pt in pts:
                f.write("%f %f %f\n" % (pt[0], pt[1], pt[2]))
            f.close()

def inverse_reg2ref(fnames):    
    """ Register the average reference to each sample to compute transformation for points """
    ref_name = "warp_ref.nrrd" 
    ref = os.path.join(final_out_dir, ref_name)
    k = 1
    for f in fnames:
        print "-->\t[",k,"/",len(fnames),"]:", f
        fname = os.path.join(data_dir, f)
        out_dir = os.path.join(inverse_dir, "reg_" + f)
        reg_name = os.path.join(reg_dir, "reg_" + f, "TransformParameters.2.txt") #
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        #cmd = "elastix -f %s -m %s -out %s -p params/rigid.txt -p params/affine.txt -p params/bspline.txt > /tmp/out.txt 2>&1" % (fname, ref, out_dir)
        cmd = "elastix -f %s -m %s -out %s -t0 %s -p params/inverse_bspline.txt > /tmp/out.txt 2>&1" % (ref, fname, out_dir, reg_name)
        print cmd
        os.system(cmd)
        k += 1

def rewrite_inv_file(fname):
    f = open(fname,'r')
    newf = open(fname+".tmp",'w')
    for l in f:
        l = l.rstrip()
        if "(InitialTransformParametersFileName" in l:
            l = "(InitialTransformParametersFileName \"NoInitialTransform\")"
        if "(Spacing" in l:
            l = "(Spacing 0.025 0.025 0.025)"
        if "(Size" in l:
            l = "(Size 528 456 320)"
        newf.write(l+"\n")
    f.close()
    os.rename(fname+".tmp", fname)

def make_count_image(pts, img_size):
    img = np.zeros(img_size)
    for p in pts:
        currp = np.round(np.array(p))
        if not np.any(currp < 0) and currp[0] < img_size[0] and currp[1] < img_size[1] and currp[2] < img_size[2]:
            img[currp[0], currp[1], currp[2]] += 1
    return img


def reg2ref(fnames, ref):
    k = 1
    for f in fnames:
        print "-->\t[",k,"/",len(fnames),"]:", f
        fname = os.path.join(data_dir, f)
        out_dir = os.path.join(base_dir, "reg", "reg_" + f)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
        cmd = "elastix -f %s -m %s -out %s -p params/rigid.txt -p params/affine.txt -p params/bspline.txt > /tmp/out.txt 2>&1" % (ref, fname, out_dir)
        print cmd
        os.system(cmd)
        k += 1

def parse_warped_points(fname):
    """ 
    Parses points of the form:
    Point   868403  ; InputIndex = [ 279 386 265 ]  ; InputPoint = [ 6.975970 9.641800 6.616390 ]   ; OutputIndexFixed = [ 272 376 226 ]    ; OutputPoint = [ 6.788555 9.391652 5.650718 ]  ; Deformation = [ -0.187415 -0.250148 -0.965672 ] 
    """
    f = open(fname,'r')
    inputidx = []
    inputpt = []
    outputidx = []
    outputpt = []
    for l in f:
        e = l.rstrip().split()
        idx = int(e[1])
        inputidx.append((int(e[6]),int(e[7]), int(e[8])))
        outputidx.append((int(e[22]), int(e[23]), int(e[24])))
        outputpt.append((float(e[30]), float(e[31]), float(e[32])))        
    f.close()
    return (inputidx, outputidx, outputpt)


